from __future__ import division

from contextlib import contextmanager
import mmap, os, struct, zlib, argparse, pynbt, io, re, itertools, sys
from cStringIO import StringIO

parser = argparse.ArgumentParser(description='mcfsck')
parser.add_argument('--entitytypes', action="store", default="ITEM",help="Comma seperated list of entitytypes to check")
parser.add_argument('--fixit', action="store_true",help="Perform write operations")
parser.add_argument('--logfile', action="store",help="Write action logs",default=sys.stdout,type=argparse.FileType('w'))
parser.add_argument('--idmap', action="store",help="Read id map and replace unknown types with air",default=None,type=argparse.FileType('r'))
parser.add_argument('files', action="store",metavar='FILE',nargs='+',help="File or folder to check")

parser.add_argument('--limitentitycount', action="store", default=500, type=int,help="Warn or delete when a group of entities is above this number")
options = parser.parse_args()

@contextmanager
def mmap_filename(filename,mode=os.O_RDONLY,access=mmap.ACCESS_READ|mmap.ACCESS_COPY):
    handle = os.open(filename,mode)
    try:
        yield mmap.mmap(handle,0,access=access)
    finally:
        os.close(handle)

mcapattern = re.compile(r"r\.([^.]+)\.([^.]+)\.mca")
def chunk_coordinates(filename,chunk_index):
    x,z = chunk_index >> 4, chunk_index & 0x1f
    result = mcapattern.match(filename)
    if result:
        region_x,region_z = map(int,result.groups())
        return "({},{})".format(x,z)
        return "({},{})".format((region_x*32+x)*16,(region_z*32+z)*16)
    return ""

def limit_entity_count(nbt):
    total_entity_count = len(nbt['Level']['Entities']) + len(nbt['Level']['TileEntities'])
    if total_entity_count > options.maxentitycount:
        options.logfile.write('{}: Chunk at {},{} {} has {} entities\n'.format(filename,nbt['Level']['xPos'].value,nbt['Level']['zPos'].value,chunk_coordinates(os.path.basename(filename),chunk_index),total_entity_count))
        nbt['Level']['Entities'][:] = [x for x in nbt['Level']['Entities'] if x['id'].value != u"Item"]
        return True

replace_blocks = {
	(90,None):(35,6),
	(119,None):(35,6),
}

import re
pattern = re.compile(r'Block\. Unused ID: (\d+)')
if options.idmap:
    for line in options.idmap:
        match = pattern.match(line)
        if match:
            replace_blocks[(int(match.group(1)),None)]=(0,0)
        

remove_block_low_set = {(block_id & 0xff) for block_id,data in replace_blocks.keys()}

import sys

def get_nible(data,index):
    if (index % 2) == 0:
        return data[index//2] & 0x0f
    else:
        return (data[index//2]>>4) & 0x0f

def set_nible(data,index, value):
    if index % 2:
        data[index//2] = (data[index//2] & 0x0f) + (value << 4)
    else:
        data[index//2] = (data[index//2] & 0xf0) + value

def get_block_replace(old_block_id,old_block_data):
    if (old_block_id,old_block_data) in replace_blocks:
        return replace_blocks[(old_block_id,old_block_data)]
    if (old_block_id,None) in replace_blocks:
        return replace_blocks[(old_block_id,None)]

non_tile_blocks = set(range(22)+[514])
def remove_air_tiles(chunk):
    modified = False
    chunk_x = chunk['Level']['xPos'].value
    chunk_z = chunk['Level']['zPos'].value
    sections = {}
    for section in chunk['Level']['Sections']:
        sections[section['Y'].value] = section
    bad_tiles = set()
    for tile in chunk['Level']['TileEntities']:
        x = tile['id'].value
	section_number = tile['y'].value//16
        block_index = (tile['x'].value & 0xf) + ((tile['y'].value & 0xf) << 8) + ((tile['z'].value & 0xf) << 4)
        block_id_low = sections[section_number]['Blocks'].value[block_index]
        block_id_high = get_nible(sections[section_number]['Add'].value,block_index) if 'Add' in sections[section_number] else 0
	block_id = block_id_low + (block_id_high << 8)
#        if x == 'factory_barrel' and ((tile['item_count'].value > 0 and 'item_type' not in tile) or (tile['item_count'].value == 0 and 'item_type' in tile)):
#        if x == 'factory_barrel':
        if x == 'factory_barrel' and tile['z'].value == 252:
#        if x == 'factory_barrel' and tile['item_count'].value > 0 and 'tag' in tile['item_type']:
            options.logfile.write("Killing {} and {}\n".format(tile.pretty(),block_id))
#            options.logfile.write("Killing {} and {}\n".format((tile['x'].value,tile['y'].value,tile['z'].value),block_id))
            bad_tiles.add(id(tile))
            sections[section_number]['Blocks'].value[block_index] = 35
            set_nible(sections[section_number]['Data'].value,block_index, 6)
            if 'Add' in sections[section_number]:
                set_nible(sections[section_number]['Add'].value,block_index,0)
	if block_id in non_tile_blocks:
            bad_tiles.add(id(tile))
    before = len(chunk['Level']['TileEntities'])
    chunk['Level']['TileEntities'][:] = [ tile for tile in chunk['Level']['TileEntities'] if id(tile) not in bad_tiles]
    after = len(chunk['Level']['TileEntities'])
    if before != after:
        options.logfile.write("Removed Tile {0} entities on air blocks\n".format(before-after))
    return before != after

def remove_block(chunk):
    modified = False
    chunk_x = chunk['Level']['xPos'].value
    chunk_z = chunk['Level']['zPos'].value
    for section in chunk['Level']['Sections']:
        section_y = section['Y'].value
        blocks = section['Blocks'].value
        add = section['Add'].value if 'Add' in section else None
        data = section['Data'].value
        for index in range(4096):
            if blocks[index] not in remove_block_low_set:
                continue
            block_y = section_y * 16 + (index >> 8)
            block_x = chunk_x   * 16 + (index & 0xf)
            block_z = chunk_z   * 16 + ((index >> 4) & 0xf)
            block_pos = (block_x, block_y, block_z)
            old_block_add = get_nible(add,index) if add else 0
            old_block_data = get_nible(data,index)
            old_block_id = blocks[index] + (old_block_add << 8)

            replacement = get_block_replace(old_block_id,old_block_data)
            if replacement:
                new_block_id,new_block_data = replacement
                new_block_id_add = new_block_id >> 8
                if new_block_id_add and 'Add' not in section:
                    options.logfile.write("Warning not replacing block at {}\n".format(block_pos))
                    continue

                blocks[index] = new_block_id & 0xff
                if 'Add' in section:
                    set_nible(add,index,new_block_id_add)
                set_nible(data,index,new_block_data)
                options.logfile.write("Replacing {} to {} @ {}\n".format((old_block_id,old_block_data),(new_block_id,new_block_data),block_pos))
                modified = True
    return modified

#fixes = [remove_air_tiles,remove_block]
#fixes = [remove_air_tiles]
#fixes = [remove_block]
fixes = [remove_block,remove_air_tiles]

def check_chunk(chunk):
    results = [func(chunk) for func in fixes]
    return any(results)

def check_file(filename):
    filename_base = os.path.basename(filename)
    with mmap_filename(filename) as data:
        file_modified = False
        for header, chunk_index in itertools.izip(struct.unpack_from(">1024i",data,0),itertools.count()):
            sys.stdout.write("\rProgress of file {} {: >3.0f}%".format(filename_base,100*chunk_index/1024))
            sys.stdout.flush()
            offset = (header >> 8) * 4096
            max_byte_size = (header & 0xFF) * 4096
            if offset:
                byte_size = struct.unpack_from('>i',data,offset)[0]
                compression = struct.unpack_from('>b',data,offset+4)[0]
                if compression != 2:
                    options.logfile.write("Warning skipping parts of file {}\n".format(filename))
                    continue
                chunk_data = zlib.decompress(buffer(data,offset+5,byte_size))
                chunk = pynbt.NBTFile(io.BytesIO(chunk_data))

                chunk_modified = check_chunk(chunk)
                if chunk_modified:
                    file_modified = True
                    new_chunk_data = StringIO()
                    chunk.save(new_chunk_data)
                    new_chunk_data_compressed = zlib.compress(new_chunk_data.getvalue(),9)
                    new_byte_size = len(new_chunk_data_compressed)
                    if new_byte_size < max_byte_size:
                        struct.pack_into('>i',data,offset,new_byte_size)
                        struct.pack_into('>b',data,offset+4,2)
                        struct.pack_into('>{}s'.format(new_byte_size),data,offset+5,new_chunk_data_compressed)
                    else:
                        options.logfile.write('Not enough room in file to fix problem max_byte_size:{}, byte_size:{}, new_byte_size:{}\n'.format(max_byte_size,byte_size,new_byte_size))
        if file_modified and options.fixit:
            with open(os.path.basename(filename) + ".fix",'w+b') as output:
                output.write(data)

    sys.stdout.write("\n".format(filename))

def main():
    for path in options.files:
        if os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for filename in files:
                    if filename.endswith('.mca'):
                        check_file(os.path.join(root,filename))
        else:
            check_file(path)

main()
